#include <stdio.h>
#include <stdlib.h>
#include "parse.h"

void cleanStmt(Stmt *);
void cleanAsgnmt(AsgnmtStmt *);
void cleanFor(ForStmt *);
void cleanWhile(WhileStmt *);
void cleanIf(IfStmt *);
void cleanCmd(CmdStmt *);
void cleanStmts(Stmts *);
void cleanWord(Word);
void cleanPiece(Piece *);
void cleanString(StringPiece *);
void cleanDouble(DoublePiece *);
void cleanVar(VarPiece *);
void cleanBack(BackPiece *);

/* cleanStmt: clean up and free a stmt.
   just calls appropriate cleaning function */
void cleanStmt(Stmt *stmt)
{
    switch (stmt->stmttype)
    {
	case ASGNMTS:
	    cleanAsgnmt((AsgnmtStmt *) stmt);
	    break;
	case FORS:
	    cleanFor((ForStmt *) stmt);
	    break;
	case WHILES:
	    cleanWhile((WhileStmt *) stmt);
	    break;
	case IFS:
	    cleanIf((IfStmt *) stmt);
	    break;
	case EOFS:
	    free(stmt);
	    break;
	default:
	    break;
    }
}

void cleanAsgnmt(AsgnmtStmt *asgnmt)
{
    free(asgnmt->var);
    cleanWord(asgnmt->val);
    free(asgnmt);
}

void cleanFor(ForStmt *forstmt)
{
    int i;
    free(forstmt->var);
    if (forstmt->vals)
	for (i = 0; forstmt->vals[i]; i++)
	    cleanWord(forstmt->vals[i]);
    free(forstmt->vals);
    cleanStmts(forstmt->body);
    free(forstmt);
}

void cleanWhile(WhileStmt *whilestmt)
{
    cleanStmts(whilestmt->cond);
    cleanStmts(whilestmt->body);
    free(whilestmt);
}

void cleanIf(IfStmt *ifstmt)
{
    cleanStmts(ifstmt->cond);
    cleanStmts(ifstmt->body);
    cleanStmts(ifstmt->elsebody);
    free(ifstmt);
}

void cleanRed(Red *red)
{
    if (red->to)
	cleanWord(red->to);
    free(red);
}

void cleanCmd(CmdStmt *cmd)
{
    int i;
    for (i = 0; cmd->args[i]; i++)
	cleanWord(cmd->args[i]);
    free(cmd->args);
    for (i = 0; cmd->reds[i]; i++)
	cleanRed(cmd->reds[i]);
    free(cmd->reds);
    if (cmd->next)
	cleanCmd(cmd->next);
    free(cmd);
}

void cleanStmts(Stmts *stmts)
{
    if (!stmts)
	return;
    cleanStmt(stmts->stmt);
    cleanStmts(stmts->next);
    free(stmts);
}

void cleanWord(Word word)
{
    int i;
    for (i = 0; word[i]; i++) {
	cleanPiece(word[i]);
    }
    free(word);
}

/* cleanPiece: clean up and free a piece.
   just calls appropriate cleaning function */
void cleanPiece(Piece *piece)
{
    switch (piece->piecetype)
    {
	case STRINGP:
	    cleanString((StringPiece *) piece);
	    break;
	case DOUBLEP:
	    cleanDouble((DoublePiece *) piece);
	    break;
	case VARP:
	    cleanVar((VarPiece *) piece);
	    break;
	case BACKP:
	    cleanBack((BackPiece *) piece);
	    break;
	default:
	    break;
    }
}

void cleanString(StringPiece *string)
{
    free(string->string);
    free(string);
}

void cleanDouble(DoublePiece *doublePiece)
{
    Piece **p;
    for (p = doublePiece->pieces; *p; p++) {
	cleanPiece(*p);
    }
    free(doublePiece);
}

void cleanVar(VarPiece *var)
{
    if (var->var)
	free(var->var);
    free(var);
}

void cleanBack(BackPiece *back)
{
    cleanStmts(back->stmts);
    free(back);
}
