#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <error.h>
#include <ctype.h>
#include "token.h"
#include "parse.h"
#include "dyninput.h"
#include "variables.h"

extern int INTERACTIVE;

AsgnmtStmt *parseAsgnmt(int *err);
CmdStmt *parseCmd(int *err);
ForStmt *parseFor(int *err);
IfStmt *parseIf(int *err);
WhileStmt *parseWhile(int *err);
Word parseWord(int *err, char *);
Red *parseRed(int *err);
VarPiece *parseVar(int *err);
StringPiece *stringPiece(char *str);
VarPiece *varPiece(char *str);
DoublePiece *parseDouble(char *str, int *err);
BackPiece *parseBack(char *str, int *err);

void unexpected(TokType tok, char *str) {
    if (tok == EOFT) {
	error(1, 0, "Unexpected end of file");
	exit(1);
    } else {
	error(0, 0, "Unexpected token '%s'", str);
    }
}

Red *parseRed(int *err)
{
    char *str;
    char *num;
    Red *red;
    TokType tok;

    red = malloc(sizeof(Red));
    
    red->tofile = TRUE;
    getToken(&str);
    num = str;
    while (*str != '<' && *str != '>')
	str++;
    /* decide redirection type */
    switch (*str) {
	case '<':
	    if (str[1] == '>') {
		red->direction = RDWR;
		free(str);
	    } else {
		red->direction = RD;
		if (str[1] == '&')
		    red->tofile = FALSE;
	    }
	    break;
	case '>':
	    if (str[1] == '>') {
		red->direction = APND;
		free(str);
	    } else {
		red->direction = WR;
		if (str[1] == '&')
		    red->tofile = FALSE;
	    }
	    break;
	default: /* this is impossible (hopefully) */
	    break;
    }
    /* set file descriptor */
    *str = '\0';
    if (num != str) 
	red->from = atoi(num);
    else /* use defaults */
	red->from = (red->direction == APND || red->direction == WR) ? 1 : 0;
    switch (tok = getToken(&str)) {
	case WORD:
	case STRINGTOKS:
	    /* set input to the word and parse it with parseWord() */
	    red->to = parseWord(err, str);
	    free(str);
	    break;
	default:
	    unexpected(tok, str);
	    free(str);
	    red->to = NULL;
	    *err = 1;
	    return red;
    }
    return red;
}

/* Stmt parsers */

/* stmtsStmt: parse next Stmts from input and return it */
Stmts *parseStmts(int body, int *err)
{
    char *str;
    Stmts *stmts;
    TokType tok;

    stmts = malloc(sizeof(Stmts));

    /* parse the first stmt */
    switch (tok = getToken(&str)) {
	case ASGNMT:
	    ungetToken(tok, &str);
	    stmts->stmt = (Stmt *) parseAsgnmt(err);
	    break;
	case WORD:
	case RED:
	    ungetToken(tok, &str);
	    stmts->stmt = (Stmt *) parseCmd(err);
	    break;
	case IF:
	    free(str);
	    stmts->stmt = (Stmt *) parseIf(err);
	    break;
	case WHILE:
	    free(str);
	    stmts->stmt = (Stmt *) parseWhile(err);
	    break;
	case FOR:
	    free(str);
	    stmts->stmt = (Stmt *) parseFor(err);
	    break;
	case EOFT:
	    /* close the list with and EOF stmt and return */
	    free(str);
	    stmts->stmt = (Stmt *) EOFStmt();
	    stmts->next = NULL;
	    return stmts;
	case NL:
	    free(str);
	    free(stmts);
	    /* if we're parsing a body, prompt for next line and get it */
	    if (body) {
		if (INTERACTIVE) {
		    fprintf(stderr, "%s", parsePrompt(getvar("PS2")));
		}
		return parseStmts(body, err);
	    }
	    /* otherwise close the list and return */
	    else 
		return NULL;
	case SEMCOL:
	    free(str);
	    free(stmts);
	    return parseStmts(body, err);
	case THEN:
	case ELSE:
	case FI:
	case DO:
	case DONE:
	    if (body) {
		/* close the list and return */
		ungetToken(tok, &str);
		free(stmts);
		return NULL;
	    }
	default:
	    *err = 1;
	    unexpected(tok, str);
	    free(str);
	    free(stmts);
	    return NULL;
    }
    if (*err) {
	stmts->next = NULL;
	return stmts;
    }
    /* parse the next stmt */
    stmts->next = parseStmts(body, err);
    return stmts;
}

/* parseAsgnmt: parse next AsgnmtStmt from input and return it */
AsgnmtStmt *parseAsgnmt(int *err)
{
    char *asgnmt;
    AsgnmtStmt *asgnmtstmt;

    asgnmtstmt = malloc(sizeof(AsgnmtStmt));
    asgnmtstmt->stmttype = ASGNMTS;
    
    getToken(&asgnmt);
    asgnmtstmt->var = asgnmt;
    /* split up the string */
    *strchr(asgnmt, '=') = '\0';
    /* find the value part of the assignment */
    asgnmt += strlen(asgnmt) + 1;
    asgnmtstmt->val = parseWord(err, asgnmt);
    return asgnmtstmt;
}

#define CHUNKSIZE 20
CmdStmt *parseCmd(int *err)
{
    CmdStmt *cmdstmt;
    TokType tok;
    int argmax = CHUNKSIZE, argc = 0;
    int redmax = CHUNKSIZE, redc = 0;
    char *str;

    cmdstmt = malloc(sizeof(CmdStmt));
    cmdstmt->stmttype = CMDS;
    
    cmdstmt->args = malloc(argmax * sizeof(Word *));
    cmdstmt->reds = malloc(redmax * sizeof(Red *));

    /* parse words until ; or \n */
    while ((tok = getToken(&str)) != SEMCOL && tok != NL && tok != PIPE && tok != EOFT && tok != AND && tok != OR) {
	switch (tok) {
 	    case WORD:
	    case STRINGTOKS:
		/* set input to current word and parse it with parseWord() */
		cmdstmt->args[argc++] = parseWord(err, str);
		free(str);
		break;
	    case RED:
		ungetToken(RED, &str);
		cmdstmt->reds[redc++] = parseRed(err);
		break;
	    default:
		unexpected(tok, str);
		free(str);
		*err = 1;
	}
	if (*err) {
	    cmdstmt->next = NULL;
	    return cmdstmt;
	}
	if (argc == argmax) {
	    argmax += CHUNKSIZE;
	    cmdstmt->args = realloc(cmdstmt->args, argmax * sizeof(Word *));
	}
	if (redc == redmax) {
	    redmax += CHUNKSIZE;
	    cmdstmt->reds = realloc(cmdstmt->reds, redmax * sizeof(Red *));
	}
    }
    cmdstmt->args[argc] = NULL;
    cmdstmt->reds[redc] = NULL;
    /* if this command ends in a newline or semicolon, unget it 
       so it can be processed elsewhere. */
    if (tok == NL || tok == SEMCOL)
	ungetToken(tok, &str);
    if (tok == PIPE || tok == AND || tok == OR) {
	cmdstmt->next = parseCmd(err);
	cmdstmt->nexttype = (tok == PIPE ? PIPEN : (tok == AND ? ANDN : ORN));
    } else {
	cmdstmt->next = NULL;
    }
    return cmdstmt;
}
#undef CHUNKSIZE


IfStmt *parseIf(int *err)
{
    IfStmt *ifstmt;
    char *str;
    TokType tok;
    
    ifstmt = malloc(sizeof(IfStmt));
    ifstmt->stmttype=IFS;
    
    ifstmt->cond = parseStmts(1, err);
    if (*err) {
	ifstmt->body = NULL;
	ifstmt->elsebody= NULL;
	return ifstmt;
    }
    if ((tok = getTokenNL(&str)) != THEN) {
	unexpected(tok, str);
	free(str);
	*err = 1;
	ifstmt->body = NULL;
	ifstmt->elsebody = NULL;
	return ifstmt;
    }
    ifstmt->body = parseStmts(1, err);
    if (*err) {
	ifstmt->elsebody = NULL;
	return ifstmt;
    }
    if ((tok = getTokenNL(&str)) != ELSE) {
	ungetToken(tok, &str);
	ifstmt->elsebody = NULL;
    } else {
	ifstmt->elsebody = parseStmts(1, err);
	if (*err)
	    return ifstmt;
    }
    if ((tok = getTokenNL(&str)) != FI) {
	unexpected(tok, str);
	free(str);
	*err = 1;
	return ifstmt;
    }
    return ifstmt;
}

WhileStmt *parseWhile(int *err)
{
    WhileStmt *whilestmt;
    char *str;
    TokType tok;
    
    whilestmt = malloc(sizeof(WhileStmt));
    whilestmt->stmttype=WHILES;
    
    whilestmt->cond = parseStmts(1, err);
    if (*err) {
	whilestmt->body = NULL;
	return whilestmt;
    }
    if ((tok = getTokenNL(&str)) != DO) {
	unexpected(tok, str);
	free(str);
	*err = 1;
	whilestmt->body = NULL;
	return whilestmt;
    }
    whilestmt->body = parseStmts(1, err);
    if (*err) {
	return whilestmt;
    }
    if ((tok = getTokenNL(&str)) != DONE) {
	unexpected(tok, str);
	free(str);
	*err = 1;
	return whilestmt;
    }
    return whilestmt;
}

#define CHUNKSIZE 20 /* amount of memory to allocate at a time for vals/args */
ForStmt *parseFor(int *err)
{
    ForStmt *forstmt;
    int valmax = CHUNKSIZE, c = 0, i;
    char *str;
    TokType tok;
    
    forstmt = malloc(sizeof(ForStmt));
    forstmt->stmttype = FORS;
    switch(tok = getToken(&str)) {
	case WORD:
	case STRINGTOKS:
	    for (i = 0; str[i]; i++) {
		if (! isalpha(str[i])) break;
	    }
	    if (i == strlen(str)) {
		forstmt->var = str;
		break;
	    }
	default:
	    fprintf(stderr, "HEY!%d", tok);
	    unexpected(tok, str);
	    free(str);
	    *err = 1;
	    forstmt->var = NULL;
	    forstmt->vals = NULL;
	    forstmt->body = NULL;
	    return forstmt;
    }
    if ((tok = getTokenNL(&str)) != IN) {
	unexpected(tok, str);
	free(str);
	*err = 1;
	forstmt->vals = NULL;
	forstmt->body = NULL;
	return forstmt;
    } else
	free(str);

    forstmt->vals = malloc(valmax * sizeof(Stmt *));
    /* parse words until ; or \n */
    while ((tok = getToken(&str)) != SEMCOL && tok != NL) {
	switch (tok) {
	    case WORD:
	    case STRINGTOKS:
		/* set input to current word and parse it with parseWord() */
		forstmt->vals[c++] = parseWord(err, str);
		free(str);
		break;
	    case EOFT:
		error(1, 0, "Unexpected end of file");
	    default:
		break;
	}
	if (c == valmax) {
	    valmax += CHUNKSIZE;
	    forstmt->vals = realloc(forstmt->vals, valmax * sizeof(Stmt *));
	}
    }
    forstmt->vals[c] = NULL;
    free(str);

    if ((tok = getToken(&str)) != DO) {
	unexpected(tok, str);
	free(str);
	*err = 1;
	forstmt->body = NULL;
	return forstmt;
    } else
	free(str);

    forstmt->body = parseStmts(1, err);
    if (*err)
	return forstmt;
    
    if ((tok = getToken(&str)) != DONE) {
	unexpected(tok, str);
	free(str);
	free(forstmt);
	return forstmt;
    } else
	free(str);
    
    return forstmt;
}
#undef CHUNKSIZE

/* EOFStmt: generate and return an EOF Stmt */
Stmt *EOFStmt()
{
    Stmt *stmt;
    
    stmt = malloc(sizeof(Stmt));
    stmt->stmttype = EOFS;
    
    return stmt;
}

/* Piece parsers */

#define CHUNKSIZE 5
Word parseWord(int *err, char *s)
{
    char *str;
    Word word;
    int i, c = 1;
    TokType tok;
    Source *bu;
    
    bu = inputString(s, strlen(s));

    word = malloc(c * sizeof(Piece *));
    
    for (i = 0; (tok = getToken2(&str)) != EOFT; i++) {
	switch (tok) {
	    case SINGLE:
		/* need to get rid of surrounding quotes */
		str[strlen(str) - 1] = '\0';
		word[i] = (Piece *) (stringPiece(strdup(str+1)));
		free(str);
		break;
	    case DOLLAR:
		free(str);
		word[i] = (Piece *) parseVar(err);
		break;
	    case TILDE:
		free(str);
		word[i] = (Piece *) varPiece("HOME");
		break;
	    case DOUBLE:
		word[i] = (Piece *) parseDouble(str, err);
		break;
	    case BACK:
		word[i] = (Piece *) parseBack(str, err);
		break;
	    case STRING:
	    default:
		word[i] = (Piece *) stringPiece(str);
		break;
	}
	if (i + 1 > c) {
	    c += CHUNKSIZE;
	    word = realloc(word, c * sizeof(Piece *));
	}
	if (*err) {
	    word[++i] = NULL;
	    restoreInput(bu);
	    return word;
	}
    }
    /* close the mofo and return */
    word[i] = NULL;
    restoreInput(bu);
    return word;
}
#undef CHUNKSIZE

/* stringPiece: generate and return StringPiece with specified string */
StringPiece *stringPiece(char *str)
{
    StringPiece *stringpiece;
    
    stringpiece = malloc(sizeof(StringPiece));
    stringpiece->piecetype = STRINGP;
    
    stringpiece->string = str;
    return stringpiece;
}

#define CHUNKSIZE 5
/* parseDouble: very similar to parseWord */
DoublePiece *parseDouble(char *dub, int *err)
{
    char *str;
    DoublePiece *doublepiece;
    TokType tok;
    int i, c = 1;
    Source *bu;
    
    doublepiece = malloc(sizeof(DoublePiece));
    doublepiece->piecetype = DOUBLEP;
    
    doublepiece->pieces = malloc(c * sizeof(Piece *));
    
    /* need to get rid of surrounding quotes */
    bu = inputString(strdup(dub+1), strlen(dub) - 2);

    for (i = 0; (tok = getToken2(&str)) != EOFT; i++) {
	switch (tok) {
	    case STRING:
	    case SINGLE:
	    case TILDE:
	    default:
		doublepiece->pieces[i] = (Piece *) stringPiece(str);
		break;
	    case DOLLAR:
		free(str);
		doublepiece->pieces[i] = (Piece *) parseVar(err);
		break;
	    case BACK:
		doublepiece->pieces[i] = (Piece *) parseBack(str, err);
		break;
	}
	if (i + 1 > c) {
	    c += CHUNKSIZE;
	    doublepiece->pieces = realloc(doublepiece->pieces, c * sizeof(Piece *));
	}
	if (*err) {
	    doublepiece->pieces[++i] = NULL;
	    restoreInput(bu);
	    free(str);
	    return doublepiece;
	}
    }
    restoreInput(bu);
    free(dub);
    return doublepiece;
}
#undef CHUNKSIZE

VarPiece *parseVar(int *err)
{
    char *str;
    VarPiece *varpiece;
    TokType tok;

    varpiece = malloc(sizeof(VarPiece));
    varpiece->piecetype = VARP;
    
    switch (tok = getToken2(&str)) {
	case STRING:
	case DOLLAR:
	    varpiece->var = str;
	    break;
	case CURLYO:
	    free(str);
	    if ((tok = getToken2(&str)) != STRING) {
		if (tok != EOFT) {
		    unexpected(tok, str);
		} else {
		    error(0, 0, "Missing variable name");
		}
		*err = 1;
		ungetToken(tok, &str);
		varpiece->var = NULL;
	    } else {
		varpiece->var = str;
		if ((tok = getToken2(&str)) != CURLYC) {
		    if (tok != EOFT) {
			unexpected(tok, str);
		    } else {
			error(0, 0, "'}' expected");
		    }
		    ungetToken(tok, &str);
		    *err = 1;
		}
	    }
	    break;
	default:
	    error(0, 0, "BAD: %s", str);
	    *err = 1;
	    ungetToken(tok, &str);
	    varpiece->var = NULL;
	    break;
    }
    return varpiece;
}

/* varPiece: generate and return a varPiece for the specified variable */
/* not sure when or why i wrote this. keeping it here just in case */
/* ha ha ha! turns out i need it! */
VarPiece *varPiece(char *str)
{
    VarPiece *varpiece;
    
    varpiece = malloc(sizeof(VarPiece));
    varpiece->piecetype = VARP;
    
    varpiece->var = str;
    return varpiece;
}

BackPiece *parseBack(char *str, int *err)
{
    BackPiece *backpiece;
    Source *bu;
    int tmp;

    backpiece = malloc(sizeof(BackPiece));
    backpiece->piecetype = BACKP;
    
    tmp = INTERACTIVE;
    INTERACTIVE = 0;

    /* need to get rid of surrounding quotes */
    bu = inputString(strdup(str+1), strlen(str) - 2);
    backpiece->stmts = parseStmts(1, err);
    restoreInput(bu);
    free(str);

    INTERACTIVE = tmp;
    
    return backpiece;
}
