#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <error.h>
#include <ctype.h>
#include "parse.h"
#include "token.h"
#include "variables.h"

extern int INTERACTIVE;

void runCmd(CmdStmt *, int);
void runFor(ForStmt *);
void runIf(IfStmt *);
void runWhile(WhileStmt *);
char **expandWords(Word *, int);
char **expandWord(Word);
void freeWords(char **);

void cleanStmts(Stmts *);

char *expandDouble(DoublePiece *doublepiece)
{
    Word words[2];
    char *ret;
    char **tmp;
    words[0] = (Word) doublepiece->pieces;
    words[1] = NULL;
    tmp = expandWords(words, 0);
    ret = strdup(tmp[0]);
    freeWords(tmp);
    return ret;
}

#define CHUNKSIZE 20 /* amount of memory to allocate at a time for buffer */

char *expandBack(BackPiece *back)
{
    int bu, pipefd[2];
    int bufmax = CHUNKSIZE;
    char *buf = malloc(sizeof(char *) * bufmax);
    int loc = 0, tmp;
    bu = dup(1);
    pipe(pipefd);
    close(1);
    dup(pipefd[1]);
    close(pipefd[1]);
    runStmts(back->stmts);
    close(1);
    while ((tmp = read(pipefd[0], buf + loc, CHUNKSIZE)) > 0) {
	loc += tmp;
	buf = realloc(buf, sizeof(char *) * (bufmax += CHUNKSIZE));
    }
    dup(bu);
    close(bu);
    return buf;
}

void runAsgnmt(AsgnmtStmt *asgnmt)
{
    char **val;

    val = expandWord(asgnmt->val);
    if (val[1]) {
	fprintf(stderr, "shitty assignment, try again!");
	return;
    }
    setvar(asgnmt->var, *val);
    freeWords (val);
}

/* expandWords: expand NULL terminated word list words[], and return
 * strings. if splitwords!=0, words are split up by spaces/newlines */
char **expandWords(Word words[], int splitwords)
{
    int i, j, k, l;
    char *tmp;
    int len = 0, c = 0; /* current total string length and word count */
    int maxwords = CHUNKSIZE; /* size of buffer currently allocated for words */
    char *buf = malloc(1); /* string buffer for all the words */
    int *expandedidxs = malloc(maxwords * sizeof(int));
    char **expanded; /* buffer for expanded words */
    
    buf[0]='\0'; // in case all the words are empty
    /* cycle through and expand all the words and stop at the end (NULL) */
    for (i = 0; words[i]; i++) {
	expandedidxs[c++] = len; /*set next string to end of buffer */
	/* cycle through and expand each piece of the word, stopping at the end (NULL) */
	for (j = 0; words[i][j]; j++) {
	    switch (words[i][j]->piecetype) {
		/* in each case, increase buffer length and copy expanded piece onto end of buffer */
		case STRINGP:
		    l = strlen(((StringPiece *) words[i][j])->string);
		    buf = realloc(buf, len + l + 1);
		    strcpy(buf + len, ((StringPiece *) words[i][j])->string);
		    break;
		case DOUBLEP:
		    tmp = expandDouble((DoublePiece *) words[i][j]);
		    l = strlen(tmp);
		    buf = realloc(buf, len + l + 1);
		    strcpy(buf + len, tmp);
		    free(tmp);
		    break;
		case VARP:
		    l = strlen(getvar(((VarPiece *) words[i][j])->var));
		    buf = realloc(buf, len + l + 1);
		    strcpy(buf + len, getvar(((VarPiece *) words[i][j])->var));
		    break;
		case BACKP:
		    tmp = expandBack((BackPiece *) words[i][j]);
		    l = strlen(tmp);
		    buf = realloc(buf, len + l + 1);
		    strcpy(buf + len, tmp);
		    free(tmp);
		    break;
	    }
	    /* do word splitting, if necessary */
	    if (splitwords && (words[i][j]->piecetype == BACKP || words[i][j]->piecetype == VARP)) {
		/* if expansion starts with a space, don't start a string here */
		if (buf[len] == ' ' || buf[len] == '\n')
		    c--;
		/* go through piece and search for spaces */
		for (k = len; buf[k]; k++) {
		    /* terminate current string at space */
		    if (buf[k] == ' ' || buf[k] == '\n') {
			buf[k] = '\0';
			/* skip additional spaces */
			while (buf[++k] == ' ' || buf[k] == '\n')
			    ;
			/* if we reach the end of the piece and there's no next piece, don't set the next string.
			   if there is another piece, set the next string provisionally; if the next piece starts
			   with a space, this string will be removed (above) */
			if (buf[k] || words[i][j + 1]) {
			    expandedidxs[c++] = k;
			    if (c >= maxwords) {
				maxwords += CHUNKSIZE;
				expandedidxs = realloc(expandedidxs, maxwords * sizeof(int));
			    }
			}
		    }
		}
	    }
	    len += l; /* augment length counter */
	}
	len++;
    }
    expanded = malloc((c+1) * sizeof(char *));
    for (i = 0; i < c; i++)
	expanded[i] = buf + expandedidxs[i];
    expanded[c] = NULL;
    free(expandedidxs);
    return expanded;
}

/* expandWord: expand a single word */
char **expandWord(Word word)
{
    Word words[2];
    words[0] = word;
    words[1] = NULL;
    return expandWords(words, 1);
}

/* freeWords: free the buffer form expandwords */
void freeWords(char **buf)
{
    free(*buf);
    free(buf);
}
    
/* runStmts: run a stmt, returns 1 if EOF was reached */
int runStmts(Stmts *stmts)
{
    if (! stmts)
	return 0;
    switch(stmts->stmt->stmttype) {
	case FORS:
	    runFor((ForStmt *) stmts->stmt);
	    break;
	case WHILES:
	    runWhile((WhileStmt *) stmts->stmt);
	    break;
	case IFS:
	    runIf((IfStmt *) stmts->stmt);
	    break;
	case CMDS:
	    runCmd((CmdStmt *) stmts->stmt, 0);
	    break;
	case ASGNMTS:
	    runAsgnmt((AsgnmtStmt *) stmts->stmt);
	    break;
	case EOFS:
	    return 1;
    }
    return runStmts(stmts->next);
}

/* runCmd: run a cmdstmt */
void runCmd(CmdStmt *cmd, int piped)
{
    char **argv;
    char **redfile;
    int i, j, tmp, bu;
    int pipefd[2];
    int stat_val;
    char retval[4];
    int child;

    /* expand word list cmd->args into char ** */
    argv = expandWords(cmd->args, 1);
    /* check for shell built-ins */
    if (! strcmp(*argv, "cd")) {
	if (argv[1]) {
	    if (chdir(argv[1])) {
		switch (errno) {
		    case EACCES:
			error(0, 0, "cd: Permission Denied");
			break;
		    case ENOENT:
			error(0, 0, "cd: %s does not exist", argv[1]);
			break;
		    case ENOTDIR:
			error(0, 0, "cd: %s is not a directory", argv[1]);
			break;
		}
	    }
	} else {
	    (chdir(getenv("HOME")));
	}
    } else if (!strcmp(*argv, "exit")) {
	exit(0);
    } else if (!strcmp(*argv, "export")) {
	for (i = 1; argv[i]; i++) {
	    if (strchr(argv[i], '=') && strrchr(argv[i], '=') == strchr(argv[i], '=')) {
		char *var, *val;
		var = val = argv[i];
		while (*++val != '=');
		*val = '\0';
		setvar(var, val+1);
		setenv(var, val+1, 1);
		*val = '=';
	    } else {
		setenv(argv[i], getvar(argv[i]), 1);
	    }
	}
    } else if (!strcmp(*argv, ".") || !strcmp(*argv, "source")) {
	bu = dup(0);
	close(0);
	if (!argv[1] || open(argv[1], O_RDONLY, 0666) == -1) {
	    if (!argv[1]) {
		error(0, 0, "filename expected!");
		errno = 0;
	    }
	    switch (errno) {
		case EACCES:
		    error(0, 0, "%s: permission denied!", argv[1]);
		    break;
		case ENOENT:
		    error(0, 0, "%s: file does not exist!", argv[1]);
		    break;
		case EISDIR:
		    error(0, 0, "%s: is a directory!", argv[1]);
		default:
		    break;
	    }
	} else {
	    /* I use i and tmp here because they're available */
	    tmp = INTERACTIVE;
	    INTERACTIVE = 0;
	    /* this is almost exactly the main loop */
	    while (1) {
		Stmts *stmts;
		i = 0;
	
		stmts = parseStmts(0, &i);
		if (! i) {
		    if (runStmts(stmts))
			break;
		} else {
		    char *str;
		    /*in case of error, chomp rest of input and give up*/
		    while (getToken(&str) != NL) 
			free(str);
		    cleanStmts(stmts);
		    break;
		}
		cleanStmts(stmts);
	    }
	    INTERACTIVE = tmp;
	}
	close(0);
	dup(bu);
	close(bu);
    } else {
	/* if not built-in, run command */
	if (cmd->next && cmd->nexttype == PIPEN)
	    pipe(pipefd);
	if (child = fork()) {
	    if (piped == 0 && INTERACTIVE) {
		piped = child;
		while (getpgid(child) == getpid());
		tcsetpgrp(0, child);
	    }
	    if (cmd->next && cmd->nexttype == PIPEN) {
		close(pipefd[1]);
		bu = dup(0);
		close(0);
		dup(pipefd[0]);
		close(pipefd[0]);
		runCmd(cmd->next, piped);
		close(0);
		dup(bu);
		close(bu);
	    } else {
		waitpid(child, &stat_val, 0);
		while (wait(NULL) != -1);
		tcsetpgrp(1, getpgid(0));
		sprintf(retval, "%d", WIFEXITED(stat_val) ?  WEXITSTATUS(stat_val) : 127);
		setvar("?", retval);
	    }
	} else {
	    if (INTERACTIVE) {
		setpgid(0, piped);
	    }
	    if (cmd->next && cmd->nexttype == PIPEN) {
		close(1);
		dup(pipefd[1]);
		close(pipefd[1]);
	    }
	    /* do redirection */
	    for (i = 0; cmd->reds[i]; i++) {
		redfile = expandWord(cmd->reds[i]->to);
		if (redfile[1]) {
		    error(0, 0, "shitty redirect, try again!");
		    exit(1);
		}
		if (cmd->reds[i]->tofile) {
		    /* I use j to hold the open() flags to avoid declaring another variable */
		    switch (cmd->reds[i]->direction) {
			case RD: j = O_RDONLY; break;
			case RDWR: j = O_RDWR; break;
			case WR: j = O_WRONLY | O_CREAT | O_TRUNC; break;
			case APND: j = O_WRONLY | O_CREAT | O_APPEND; break;
			default: exit(1); break;
		    }
		    close(cmd->reds[i]->from);
		    if (open(redfile[0], j, 0666) == -1) {
			switch (errno) {
			    case EACCES:
				error(0, 0, "%s: permission denied!", redfile[0]);
				break;
			    case ENOENT:
				error(0, 0, "%s: file does not exist!", redfile[0]);
				break;
			    case EISDIR:
				error(0, 0, "%s: is a directory!", redfile[0]);
			    default:
				break;
			}
			exit(1);
		    }
		} else if (!strcmp(redfile[0], "-")) {
		    close(cmd->reds[i]->from);
		} else {
		    for (j = 0; redfile[0][j]; j++) {
			if (!isdigit(redfile[0][j])) {
			    error(0, 0, "shitty redirect, try again!");
			    exit(1);
			}
		    }
		    close((cmd->reds[i]->direction == WR) ? cmd->reds[i]->from : atoi(redfile[0]));
		    dup((cmd->reds[i]->direction == RD) ? cmd->reds[i]->from : atoi(redfile[0]));
		}
	    }
	    /* ignore stops, because no job control */
	    signal(SIGTSTP, SIG_IGN);
	    /* set terminal block signals back to normal */
	    signal(SIGTTOU, SIG_DFL);
	    signal(SIGTTIN, SIG_DFL);
	    execvp(argv[0], argv);
	    switch(errno) {
		case ENOENT:
		    error(0, 0, "%s: command not found", argv[0]);
		    break;
		case EACCES:
		    error(0, 0, "%s: Permission denied", argv[0]);
		    break;
		default:
		    break;
	    }
	    /* terminate and kill everything else in pipeline */
	    kill(0, SIGKILL);
	} /* if we are the parent */
	/* run the "and" or "or" command if there is one */
	if (cmd->next && cmd->nexttype == (!atoi(getvar("?")) ? ANDN : ORN)) {
	    runCmd(cmd->next, 0);
	}
    } /* if command wasn't built in */
    freeWords(argv);
}

void runFor(ForStmt *forstmt)
{
    char **vals;
    int i;

    /* expand word list forstmt->vals into char ** */
    vals = expandWords(forstmt->vals, 1);
    for (i = 0; vals[i]; i++) {
	setvar(forstmt->var, vals[i]);
	runStmts(forstmt->body);
    }
    freeWords(vals);
}

void runIf(IfStmt *ifstmt)
{
//    runCmd(ifstmt->cond);
    runStmts(ifstmt->cond);
    if (atoi(getvar("?")) == 0)
	runStmts(ifstmt->body);
    else if (ifstmt->elsebody)
	runStmts(ifstmt->elsebody);
}

void runWhile(WhileStmt *whilestmt)
{
    while (runStmts(whilestmt->cond), atoi(getvar("?")) == 0)
	runStmts(whilestmt->body);
}
