#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include "token.h"
#include "dyninput.h"

extern int INTERACTIVE;

#define CHUNKSIZE 20 /* amount of memory to allocate at a time for buffer */

/* pushback storage */
static TokType backtok;
static char *backstr = NULL;

/* getToken: get a token from stdin or from pushback */
TokType getToken(char **str)
{
    int c, i;
    int strsize = 0; /* current size of string */
    int strmax = CHUNKSIZE; /* size of buffer currently allocated for string */
    /* QUOTECHAR: type of quote in which getToken is currently
     * initialized to 0 for lack of a better value */
    enum { NONE = 0, DOUBLE = '"', SINGLE = '\'', BACK = '`'} QUOTECHAR = NONE;
    
    /* check for pushed back token */
    if (backstr) {
	*str = backstr;
	backstr = NULL;
	return backtok;
    }

    /* allocate initial memory for string */
    *str = malloc(strmax*sizeof(char));

    /* skip white space */
    while ((c = getch()) == ' ')
	;
    if (c == EOF)
	return EOFT;
    /* set first character of string and check for special cases */
    (*str)[strsize++] = c;
    (*str)[strsize] = '\0';
    switch (c) {
	case ';' : return SEMCOL;
	case '|' : 
	    if ((c = getch()) == '|') {
		return OR;
	    } else {
		return PIPE;
	    }
	case '\n': return NL;
	case '"' : QUOTECHAR = DOUBLE; break;
	case '\'': QUOTECHAR = SINGLE; break;
	case '`' : QUOTECHAR = BACK;   break;
	case '>' :
	case '<' :
	    if ((c = getch()) == '>' || c == '&') {
		(*str)[strsize++] = c;
		/* make room for the extra character if necessary */
		if (strsize == strmax)
		    *str = realloc(*str, (++strmax)*sizeof(char));
		(*str)[strsize] = '\0';
	    } else
		assert(ungetch(c));
	    return RED;
	case '&' : 
	    if ((c = getch()) == '&') {
		return AND;
	    } else {
		assert(ungetch(c));
	    }
	    break;
	case '#':
	    (*str)[strsize - 1] = '\n';
	    while (getchar() != '\n');
	    return NL;
	case '\\':
	    strsize--; break;
	default: break;
    }

    /* get the rest of the characters */
    while ((c = getch()) != EOF &&
	   (QUOTECHAR == NONE 
	    ? c!=' ' && c != '\n' && c != ';' && c != '|' && c != '<' && c!= '>' && c != '&'
	    : 1)) {
	if (strsize == strmax) {
	    strmax += CHUNKSIZE;
	    *str = realloc(*str, strmax*sizeof(char));
	}
	switch (c) {
	    case '\\':
		if (QUOTECHAR == SINGLE)  {
		    (*str)[strsize++] = '\\';
		    break;
		}
		if ((c = getch()) == '\n') continue;
		break;
	    case '"' :
		if (QUOTECHAR == DOUBLE) QUOTECHAR = NONE;
		else if (QUOTECHAR == NONE) QUOTECHAR = DOUBLE;
		break;
	    case '\'':
		if (QUOTECHAR == SINGLE) QUOTECHAR = NONE;
		else if (QUOTECHAR == NONE) QUOTECHAR = SINGLE;
		break;
	    case '`' :
		if (QUOTECHAR == BACK) QUOTECHAR = NONE;
		else if (QUOTECHAR == NONE) QUOTECHAR = BACK;
		break;
	    case '#':
		while (getchar() != '\n');
		ungetch('\n');
		continue;
	    case '\n':
		if (INTERACTIVE && QUOTECHAR != NONE) {
		    fprintf(stderr, "%s", parsePrompt(getvar("PS2")));
		}
		break;
	    default: break;
	}
	(*str)[strsize++] = c;
    }
    if (strsize == strmax) {
	strmax++;
	*str = realloc(*str, strmax*sizeof(char));
    }

    /* handle super duper special case of redirections */
    if (c == '<' || c == '>') {
	for (i = 0; i < strsize && isdigit((*str)[i]); i++);
	if (i == strsize) {
	    (*str)[strsize++] = c;
	    /* make room for the extra 1 or two characters if necessary */
	    if (strsize == strmax)
		*str = realloc(*str, (strmax+=2)*sizeof(char));
	    (*str)[strsize] = '\0';
	    if ((c = getch()) == '>' || c == '&') {
		(*str)[strsize++] = c;
		(*str)[strsize] = '\0';
	    } else
		assert(ungetch(c));
	    return RED;
	}
    }

    /* push back extra character and terminate string */
    if (c != EOF)
	ungetch(c);
    else if (QUOTECHAR != NONE) {
	error(1, 0, "Unexpected EOF while searching for matching `%c'", QUOTECHAR);
    }
    (*str)[strsize] = '\0';

    if (strchr(*str, '=')) {
	for (i = 0; (*str)[i] != '='; i++) {
	    if (! isalnum((*str)[i])) break;
	}
	if(!isdigit(**str) && (*str)[i] == '=') {
	    return ASGNMT;
	}
    }
    if (!strcmp(*str, "for"))
	return FOR;
    if (!strcmp(*str, "in"))
	return IN;
    if (!strcmp(*str, "while"))
	return WHILE;
    if (!strcmp(*str, "do"))
	return DO;
    if (!strcmp(*str, "done"))
	return DONE;
    if (!strcmp(*str, "if"))
	return IF;
    if (!strcmp(*str, "then"))
	return THEN;
    if (!strcmp(*str, "else"))
	return ELSE;
    if (!strcmp(*str, "fi"))
	return FI;
    return WORD;
}

/* getTokenNL: get a token, but skip newlines and prompt if user-input */
TokType getTokenNL(char **str){
    TokType tok;
    while ((tok = getToken(str)) == NL) {
	if (INTERACTIVE) {
	    fprintf(stderr, "%s", parsePrompt(getvar("PS2")));
	}
	free(*str);
    }
    return tok;
}

/* ungetToken: pushes back a single token  */
int ungetToken(TokType tok, char **str)
{
    if (backstr)
	return -1;
    else {
	backstr = *str;
	backtok = tok;
	return 0;
    }
}

/* getToken2: second pass tokenizer for strings */
/* basically a stripped down version of getToken with some changes */
TokType getToken2(char **str)
{
    int c;
    int strsize = 0; /* current size of string */
    int strmax = CHUNKSIZE; /* size of buffer currently allocated for string */
    /* QUOTEBACK: type of quote in which getToken2 is currently */
    /* initialized to 0 for lack of a better value */
    /* DOUBLE is excluded because these are returned as tokens (for use with parseDouble) */
    enum { Q_NONE = 0, Q_SINGLE = '\'', Q_BACK = '`', Q_DOUBLE = '"' } QUOTECHAR = Q_NONE; 
    int i; /* iterator, used only in case of redirect */
    
    /* check for pushed back token */
    if (backstr) {
	*str = backstr;
	backstr = NULL;
	return backtok;
    }

    /* allocate initial memory for string */
    *str = malloc(strmax*sizeof(char));

    /* set first character of string and check for special cases */
    (*str)[strsize++] = c = getch();
    (*str)[strsize] = '\0';
    switch (c) {
	case EOF : return EOFT;
	case '\\': break;
	case '$' : 
	    if ((c = getch()) == '=') {
		(*str)[strsize++] = c;
	    }  else if (c == '\'' || c == '"' || c == '`') {
		ungetch(c);
		return STRING;
	    } else if (c != EOF) {
		ungetch(c);
		return DOLLAR;
	    }
	    break;
	case '{' : return CURLYO;
	case '}' : return CURLYC;
	case '"' : QUOTECHAR = Q_DOUBLE; break;
	case '`' : QUOTECHAR = Q_BACK  ; break;
	case '\'': QUOTECHAR = Q_SINGLE; break;
	case '~':
	    if ((c = getch()) == ' ' || c == '/' || c == EOF) {
		if (c != EOF) ungetch(c);
		return TILDE;
	    }
	    ungetch(c);
	    break;
	default  : break;
    }

    /* get the rest of the characters */
    while ((c = getch()) != EOF &&
	   (QUOTECHAR == Q_DOUBLE
	    ? c!= '"'
	    :(QUOTECHAR == Q_BACK
	      ? c != '`'
	      : (QUOTECHAR == Q_SINGLE
		 ? c != '\''
		 : c != '$' && c != '}' && c != '`' && c!= '\'' && c != '"')))) {	       
	if (c == '\\')
	    c = getch();
	(*str)[strsize++] = c;
	if (strsize == strmax) {
	    strmax += CHUNKSIZE;
	    *str = realloc(*str, strmax*sizeof(char));
	}
    }

    /* push back extra character (or add it if quote) and terminate string*/
    if ((c == '\'' && QUOTECHAR == Q_SINGLE) || (c == '`' && QUOTECHAR == Q_BACK) || (c == '"' && QUOTECHAR == Q_DOUBLE)) {
	/* make room for the extra characters if necessary */
	if (strsize == strmax)
	    *str = realloc(*str, (++strmax)*sizeof(char));
	(*str)[strsize++] = c;
    } else if (c != EOF)
	ungetch(c);
    (*str)[strsize] = '\0';

    switch (QUOTECHAR) {
	case Q_DOUBLE: return DOUBLE;
	case Q_BACK : return BACK;
	case Q_SINGLE: return SINGLE;
	default  : return STRING;
    }	    
}
